/*******************************************************************************
* Copyright 2019-2021 Intel Corporation
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*******************************************************************************/

#ifndef GPU_GPU_MATMUL_PD_HPP
#define GPU_GPU_MATMUL_PD_HPP

#include <assert.h>

#include "common/c_types_map.hpp"
#include "common/matmul_pd.hpp"

namespace dnnl {
namespace impl {
namespace gpu {

struct gpu_matmul_pd_t : public matmul_pd_t {
    using matmul_pd_t::matmul_pd_t;

    bool has_blocks() {
        for (auto md : {&src_md_, &weights_md_, &bias_md_, &dst_md_}) {
            memory_desc_wrapper mdw(md);
            if (mdw.is_blocking_desc()) {
                if (mdw.blocking_desc().inner_nblks != 0) { return true; }
            }
        }
        return false;
    }
};

} // namespace gpu
} // namespace impl
} // namespace dnnl

#endif
