/*******************************************************************************
* Copyright 2018-2020 Intel Corporation
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*******************************************************************************/

#include <stdio.h>
#include "dnnl.h"

int main() {
    printf("dnnl_version: %d.%d.%d\n", DNNL_VERSION_MAJOR, DNNL_VERSION_MINOR,
            DNNL_VERSION_PATCH);
    printf("dnnl_memory_desc_init_by_tag = %p, "
           "sizeof(dnnl_memory_desc_t) = %d\n",
            dnnl_memory_desc_init_by_tag, (int)sizeof(dnnl_memory_desc_t));
    return 0;
}
